#!/usr/bin/perl
# NickSentry by rys <rys@aliens.com>
# Copyright(c)1999. All Rights Reserved.

use IO::Socket;
use IO::Select;

# Configuration
$nsversion = "Fri May 21 09:45:47 CDT 1999";
$botnick = "primary"; # Primary Nick
$password = "eln5XKlbKFF4c"; # Password crypt/plain text.
$cryptpwd = 1; # Password is crypted.
$delay = 4; # Delay
$joinchan = undef; # Channel to join on startup or undef
# Log features are not quite implemented yet.
$logfile = "~/.sentrylog"; # File in which to log message or undef for none
$keeplog = 1; # Keep log after exit
$reconnect = -1; # Times to attempt to reconnect; -1 for infinite.
$emailinfo = "Nick Sentry"; # Email info
$hostname = "hostname";
$ircserver = "irc.idle.net";
$ircport = "6667";
$idletime = time;

# CTCP Responses

my %clientinfo = (action=>"ACTION contains action descriptions for atmosphere",
bdcc=>"BDCC checks cdcc info for you",
cdcc=>"CDCC checks cdcc info for you",
clientinfo=>"CLIENTINFO gives information about available CTCP commands",
dcc=>"DCC requests a direct_client_connection",
echo=>"ECHO returns the arguments it receives",
errmsg=>"ERRMSG returns error messages",
finger=>"FINGER shows real name, login name and idle time of user",
ident=>"IDENT change userhost of userlist",
invite=>"INVITE invite to channel specified",
op=>"OP ops the person if on userlist",
ops=>"OPS ops the person if on userlist",
ping=>"PING returns the arguments it receives",
sed=>"SED contains simple_encrypted_data",
time=>"TIME tells you the time on the user's host",
unban=>"UNBAN unbans the person from channel",
uptime=>"UPTIME my uptime",
userinfo=>"USERINFO returns user settable information",
utc=>"UTC substitutes the local timezone",
version=>"VERSION shows client type, version and environment",
whoami=>"WHOAMI user list information",
xdcc=>"XDCC checks cdcc info for you",
xlink=>"XLINK x-filez rule");

# CTCP VERSION Replies
$version = "\02BitchX-75p1+\02 by panasync \02-\02 OpenBSD 2.4 :\02 Keep it to yourself!\02";
$clientinfo = "SED UTC ACTION DCC CDCC BDCC XDCC VERSION CLIENTINFO USERINFO ERRMSG FINGER TIME PING ECHO INVITE WHOAMI OP OPS UNBAN IDENT XLINK UPTIME  :Use CLIENTINFO <COMMAND> to get more specific information";

# Sub routines

sub CTCP {
# Define locals
my $ctcp_from; my $ctcp_cmd; my $ctcp_to; my $ctcp_msg; my $ctcp_type;
my @ctcp_parsemsg;

# Divide the line
($ctcp_from,$ctcp_cmd,$ctcp_to,@ctcp_parsemsg) = split(/\s/,$_[0]);
$ctcp_msg = join(' ',@ctcp_parsemsg);

# If the message isn't a PRIVMSG then there's nothing more we can do
if ($ctcp_cmd ne "PRIVMSG") { return undef; }

# If the message is to the channel don't respond to it
if ($ctcp_to =~ /^\#/) { return undef; }

# Strip it apart. As long as the message part of the line starts and ends
# with ^A then grab the line contents, otherwise it's not a ctcp, so return
if ($ctcp_msg =~ /:\01(.*)\01$/) { $ctcp_type=$1; } else { return undef; }

# If it is an action then this isn't for us either
if ($ctcp_type =~ /^ACTION/i) { return undef; }

# If we're still continuing, let's get the user's name so we can reply
# if there are exclamation points, get the nick and reply like any other client
# remove preceeding colons
$ctcp_from =~ s/^\://; if ($ctcp_from =~ /^(.*)\!/) { $ctcp_from = $1; }

if ($ctcp_type =~ /^VERSION$/)
    { return("NOTICE $ctcp_from :\01$ctcp_type ".$version."\01\n"); }

if ($ctcp_type =~ /^CLIENTINFO/)
    { $ctcp_type =~ /^CLIENTINFO(.*)$/;
    my $clientinfo_type; my $clientinfo_type_disp;
    if ($1)
        { $clientinfo_type_disp = $1; }
    else
        { return("NOTICE $ctcp_from :\01CLIENTINFO ".$clientinfo."\01\n"); }
    $clientinfo_type_disp=~s/^\s//;
    ($clientinfo_type=$clientinfo_type_disp)=~tr/A-Z/a-z/;

    if(defined($clientinfo{$clientinfo_type}))
        { return("NOTICE $ctcp_from :\01CLIENTINFO ".$clientinfo{$clientinfo_type}."\01\n"); }
    else
        { return("NOTICE $ctcp_from :\01ERRMSG CLIENTINFO: $clientinfo_type_disp is not a valid function\01\n"); } }

if ($ctcp_type =~ /^PING/)
    { return("NOTICE $ctcp_from $ctcp_msg\n"); }

if ($ctcp_type =~ s/^ERRMSG//)
    { return("NOTICE $ctcp_from $ctcp_msg\n"); }

if ($ctcp_type =~ /^FINGER$/)
    { my $idlesecs = (time-$idletime); my $idleplural = "s";
    if ($idlesecs == 1)
        { undef $idleplural; }
    return("NOTICE $ctcp_from :\01FINGER $emailinfo ($botnick\@$hostname) Idle ".$idlesecs." second".$idleplural."\01\n"); }

if ($ctcp_type =~ /^OP|OPS/)
    { my $ctcp_opchan;
    if(!($ctcp_opchan = (split(/\s+/,$ctcp_type))[1]))
        { return undef; }
    return("NOTICE $ctcp_from :\02BitchX\02: I'm not on $ctcp_opchan, or I'm not opped\n"); }

if ($ctcp_type =~ /^UNBAN/)
    { my $ctcp_banchan;
    if(!($ctcp_banchan = (split(/\s+/,$ctcp_type))[1]))
        { return undef; }
	#
	# If banchan == channel that we are on return "Access Denied"
	# 
    return("NOTICE $ctcp_from :\02BitchX\02: I'm not on that channel\n"); }

if ($ctcp_type =~ /^INVITE/)
    { my $ctcp_invchan;
    if(!($ctcp_invchan = (split(/\s+/,$ctcp_type))[1]))
        { return undef; }
	#
	# If invchan == channel that we are on return "Access Denied"
	# 
    if ($ctcp_invchan !~ /^\#/)
        { return undef; }
    return("NOTICE $ctcp_from :\02BitchX\02: I'm not on that channel\n"); }

if ($ctcp_type =~ /^WHOAMI/)
    { return("NOTICE $ctcp_from :\02BitchX\02: Access Denied\n"); }

if ($ctcp_type =~ s/^UTC\s//)
    { $ctcp_type =~ s/\s+/ /g; my $ctcp_time;
    if (/\s/)
        { $ctcp_time =~ (split(/\s/,$ctcp_type))[0]; }
    else
        { $ctcp_time = $ctcp_type; }
    return("PRIVMSG $ctcp_from :".localtime($ctcp_time)."\n"); }

return undef;
}

sub Checksum {
my $cs_filename = glob($0);
my @cs_binpaths = ("/bin","/sbin","/usr/bin","/usr/sbin",
                  "/usr/local/bin","/usr/local/sbin");
unless(-f $cs_filename) { return undef; }
foreach(@cs_binpaths) {
if (-x "$_/md5") { return((split(/\s+/,`$_/md5 $cs_filename`))[3]); } }
return undef;
}

sub SocketOUT { syswrite($main_socket, $_[0], length($_[0])); return 1; }

sub ConnectTCP {
$main_socket = new IO::Socket::INET(PeerAddr=>$ircserver,PeerPort=>$ircport,Proto=>'tcp');
&LogRAW("Attempting to connect to ".$ircserver.":".$ircport."[".($attempt+1)."]");
unless($main_socket) { unless($running) { print("Error connecting: ".$!."\n"); }
                       &LogRAW("Error connecting: ".$!); return undef; }
$readable_handles->add($main_socket);
&SocketOUT("USER $botnick +iw $botnick :$emailinfo\n");
&SocketOUT("NICK ".$botnick."\n"); undef $registered; undef $attempt;
&LogRAW("Connected to server");
return 1;
}

sub ObtainNick { if ($disabled < time) { &SocketOUT("NICK ".$botnick."\n"); $disabled = time+($delay-1); } }

sub GenerateAltNick { local($gantext,$ganlp);
for($ganlp=0;$ganlp!=8;$ganlp++) { $gantext .= chr(rand(26)+65); } 
return $gantext;
}

sub LogMessage {
return undef unless defined($_[0])&&defined($_[1])&&defined($_[2])&&defined($logfile);

$_[0] =~ /^\:([^\!]+)\!([^\@]+)\@(.+)$/;
my $lm_nick = $1; my $lm_user = $2; my $lm_addy = $3; my $lm_cmd = $_[1];
my $lm_ctcp; my $lm_type; my $lm_msg = $_[2]; $lm_msg =~ s/(\x0d|\x0a|\r|\n)//g;
my $tstamp = localtime(time);
if ($lm_msg =~ /^\01(.*)\01$/) {
    $lm_ctcp = $1;
    if ($lm_cmd =~ /^NOTICE$/i)
        { if ($lm_ctcp =~ /\s/)
              { $lm_ctcp =~ /([^\ ]+)*\ *(.+)/;
                $lm_type = $1."_REPLY from ".$lm_nick; $lm_msg = $2; }
           else
              { $lm_type = $lm_ctcp; }
        }
    if ($lm_cmd =~ /^PRIVMSG$/)
        { if ($lm_ctcp =~ /^ACTION*(.+)$/)
              { $lm_type = "ACTION"; $lm_msg = $lm_nick.$1; }
          else
              { $lm_type = "CTCP_REQUEST from ".$lm_nick;
                ($lm_msg = $lm_ctcp) =~ /([^\ ]+)/; $lm_msg = $1; }
        }
}
if (!$lm_type) { $lm_type = "MSG"; $lm_msg = "*".$lm_nick."* ".$lm_msg; }
$lm_type =~ tr/a-z/A-Z/;
unless(open(LOG_FD,">> ".$logfile)) { return undef; }
print(LOG_FD "[".$tstamp."] ".$lm_type.": ".$lm_msg."  (".$lm_user."\@".$lm_addy.")\n"); $messages++;
close LOG_FD;
return 1;
}

sub LogCMD {
# $_[0] is user
# $_[1] is the command and parameters to log
return undef unless defined($_[0])&&defined($_[1])&&defined($logfile);
my $tstamp = localtime(time);
my $lc_user = $_[0]; $lc_user =~ s/^\://;
unless(open(LOG_FD,">> ".$logfile)) { return undef; }
print(LOG_FD "[".$tstamp."] # ".$_[1]."  (".$lc_user.")\n"); $messages++;
close LOG_FD;
return 1;
}

sub LogRAW {
# $_[0] is message to log
return undef unless defined($_[0])&&defined($logfile);
my $tstamp = localtime(time);
unless(open(LOG_FD,">> ".$logfile)) { return undef; }
print(LOG_FD "[".$tstamp."] ".$_[0]."\n"); $messages++;
close LOG_FD;
return 1;
}

#^ADCC CHAT chat 2130706433 14025^A
sub LogHandler {
if ((!defined($_[0]))&&(defined($dcchandle)))
    { # dump log
    }

}

# Main

print "NickSentry by rys <rys\@aliens.com>\n";
print "Copyright(c)1999. All Rights Reserved.\n";
print "Version: ".$nsversion."\n\n";

if ($ARGV[0]) { $ircserver=$ARGV[0];
                print("Using server: $ircserver\n"); }
if ($logfile) { $logfile = glob($logfile); 
                print("Using logfile: $logfile\n"); }
if (($ARGV[0])||($logfile)) { print("\n"); }

&LogRAW("NickSentry Startup");

$readable_handles = new IO::Select(); 

initconn:while(($attempt < 5)&&(!$main_socket))
{ if (&ConnectTCP) { $disabled = time + 20; last initconn; }
  $attempt++;
  sleep 10; }
unless($main_socket) { exit 1; }

$emailinfo = " " unless $emailinfo;
selectloop:while(1) {
    ($new_readable) = IO::Select->select($readable_handles,undef,undef,$delay);
    foreach $sock (@$new_readable)
        { $buf = <$sock>;
          if ($buf) {
             my $resp; if ($resp = &CTCP($buf)) { &SocketOUT($resp); }
             my @splitbuf = split(/\s+/,$buf);
             if ($buf =~ /^ERROR \:(.+)/) { LogRAW("Error from server: $1"); $error++; $servererrors++; if ($servererrors > 5) { print("Too Many Errors: ".$1."\n"); LogRAW("Too many errors. Exitting..."); exit 1; } }
             if ($buf =~ /^PING :(.+)$/) { &SocketOUT("PONG :$1\n"); }
             if ($splitbuf[1] eq "001") { unless($running) { print("Started running with nick: ".$splitbuf[2]."\n") } LogRAW("Connected to server with nick: ".$splitbuf[2]); $running++; $registered++; $disabled = time + 3; if($joinchan) { &SocketOUT("JOIN ".$joinchan."\n"); } }
             if (($splitbuf[1] eq "433")&&(!$registered))
                  { &SocketOUT("NICK ".&GenerateAltNick."\n"); }
             if ($splitbuf[1] eq "NOTICE")
                  { my $linetext = (split(/\:/,$buf))[2];
                    &LogMessage($splitbuf[0],$splitbuf[1],$linetext); }
             if ($splitbuf[1] eq "PRIVMSG")
                  { my $linetext = (split(/\:/,$buf))[2];
                    my @linearry = split(/\s+/,$linetext);
                    my $resptarg = $splitbuf[0]; 
                    $resptarg =~ /\:([^\!]+)\!/; $resptarg = $1;
                    my $userpwd = $linearry[1];
                    if ($cryptpwd) { $userpwd = crypt($userpwd,$password); }
                    unless($userpwd eq $password) {
                       if ($splitbuf[2] !~ /^\#/)
                          { &LogMessage($splitbuf[0],$splitbuf[1],$linetext); }
                    next selectloop; }
                    if ($linearry[0] =~ /^version$/i) { 
                        &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Version: ".$nsversion."\n"); my $md5 = &Checksum; if ($md5) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] MD5 Checksum: ".$md5."\n"); } $idletime = time; &LogCMD($splitbuf[0],"version"); next selectloop; }
                    if ($linearry[0] =~ /^release$/i) { 
                    $disabled = time + 45; &SocketOUT("NICK ".&GenerateAltNick."\n"); 
                    &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Giving up nick for 45 seconds.\n"); $idletime = time; &LogCMD($splitbuf[0],"release"); next selectloop; }
                    if ($linearry[0] =~ /^join$/i) {
                       unless($linearry[2]) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Usage: join <password> <channel>\n"); $idletime = time; next selectloop; }
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Attempting to join: ".$linearry[2]."\n"); $idletime = time;
                       &SocketOUT("JOIN ".$linearry[2]."\n");
                       &LogCMD($splitbuf[0],"join ".$linearry[2]);
                       next selectloop; }
                    if ($linearry[0] =~ /^part$/i) {
                       unless($linearry[2]) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Usage: part <password> <channel>\n"); $idletime = time; next selectloop; }
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Parting: ".$linearry[2]."\n"); $idletime = time;
                       &SocketOUT("PART ".$linearry[2]."\n");
                       &LogCMD($splitbuf[0],"part ".$linearry[2]);
                       next selectloop; }
                    if ($linearry[0] =~ /^delay$/i) {
                       unless($linearry[2] =~ /\d/) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Usage: delay <password> <delay>\n");
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Current delay: $delay\n");
                       $idletime = time; next selectloop; }
                       $delay = $linearry[2]; 
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Delay set to: $delay\n");
                       &LogCMD($splitbuf[0],"delay ".$delay);
                       $idletime = time; next selectloop; }
                    if ($linearry[0] =~ /^log$/i) { &LogHandler(@linearry); next selectloop; }
                    if ($linearry[0] =~ /^nick$/i) {
                       unless($linearry[2]) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Usage: nick <password> <nick>\n"); $idletime = time; next selectloop; }
                       &SocketOUT("NICK ".$linearry[2]."\n");
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Attempting to change current nick to: ".$linearry[2]."\n"); $idletime = time; &LogCMD($splitbuf[0],"nick ".$linearry[2]); next selectloop; }
                    if ($linearry[0] =~ /^orignick$/i) {
                       unless($linearry[2]) { &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Usage: orignick <password> <nick>\n");
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] Currently capturing: $botnick\n");
                       $idletime = time; next selectloop; }
                       $botnick = $linearry[2]; 
                       &SocketOUT("PRIVMSG ".$resptarg." :[NickSentry] OrigNick on: $botnick\n");
                       $idletime = time; &LogCMD($splitbuf[0],"orignick ".$botnick); next selectloop; }
                    if ($linearry[0] =~ /^(die|restart)$/i) {
                       my $quitmsg; my $dr_type = $1;
                       if ($linearry[2]) { shift @linearry; shift @linearry;
                       $quitmsg = join(' ',@linearry); }
                       unless($quitmsg) {
                        if ($dr_type =~ /^die$/i) { $quitmsg = "Shutdown"; }
                        if ($dr_type =~ /^restart$/i) { $quitmsg = "Restart"; }
                       $quitmsg .= " Requested"; }
                       &LogCMD($splitbuf[0],$dr_type);
                       &SocketOUT("QUIT :".$quitmsg."\n");
                       if ($dr_type =~ /^die$/i) { last selectloop; }
                       }
                    }
             next selectloop;
          } else {
          $readable_handles->remove($sock); LogRAW("Disconnected from server");
          undef $main_socket;
          close($sock); 
          recon:while(($attempt < $reconnect)||($reconnect < 0)&&(!$main_socket))
	  { if (&ConnectTCP) { $disabled = time + 20; last recon; }
            $attempt++; sleep 10; }
          unless ($main_socket) { exit 1; }
          }
        }
        # Event loop
        &ObtainNick;
}
if ((-f $logfile)&&(!$keeplog)) { unlink($logfile); }
&LogRAW("NickSentry Shutdown");
exit 1;

